var pdfUrl = "__gtb_pdf_url__";
var lastPageNum = __gtb_last_page_num__;
var renderingHintText = "__gtb_rendering_hint_text__";
var renderingFailText = "__gtb_rendering_fail_text__";

var pageScaleLimit = 4.0;
var desiredWidth = screen.width;

var pdfDoc = null;
var pdfPages = null;
var pdfPagesRenderScale = null;
var pdfPagesRenderState = null;

var RENDER_STATE_INIT = 0;
var RENDER_STATE_PARSED = 1;
var RENDER_STATE_RENDERING = 2;
var RENDER_STATE_RENDERED = 3;

var numParsedPage = 0;
var currentPageNum = 0;
var firstPagePaint = false;
var willRenderPageNumLow = 0;
var willRenderPageNumHigh = 0;

var loadingPercent = 0;
var renderingPercent = 0;
var percentWeight = 0.6;
var fakeLoadingInterval = 50;
var fakeLoadingTimeoutId;

var loadingFailTimeout = 20000;
var loadingFailTimeoutId;
var loadingFailed = false;

LoadPDFFile();

function LoadPDFFile() {
	document.getElementById("download_hint").innerText = renderingHintText;
	loadingFailTimeoutId = setTimeout(function(){
		if (!firstPagePaint) {
			var downloadPercent = document.getElementById("download_percent");
			downloadPercent.innerText = "0%";
			downloadPercent.style.color = "#ababab";
			var downloadHint = document.getElementById("download_hint");
			downloadHint.innerText = renderingFailText;
			downloadHint.style.color = "#ababab";
			var loadingImg = document.getElementById("loading_img");
			loadingImg.src = "./loading_fail.png";
			loadingImg.style.animationPlayState = "paused";
			clearTimeout(fakeLoadingTimeoutId);
			loadingFailed = true;
		}
	}, loadingFailTimeout);
	
	PDFJS.cMapUrl = './external/bcmaps/';
	PDFJS.cMapPacked = true;
	var loadingTask = PDFJS.getDocument(pdfUrl, null, null, function(progress) {
		if (loadingFailed) return;
		loadingPercent = Math.floor(100.0 * progress.loaded / progress.total);
		var downloadPercent = document.getElementById("download_percent");
		downloadPercent.innerText = Math.floor(loadingPercent * percentWeight + renderingPercent * (1.0 - percentWeight)) + "%";
		if (progress.loaded == progress.total) {
			fakeLoadingTimeoutId = setTimeout(FakeLoading, fakeLoadingInterval);
		}
	});
	loadingTask.then(function(doc) {
		pdfDoc = doc;
		pdfPages = new Array(pdfDoc.numPages);
		pdfPagesRenderScale = new Array(pdfDoc.numPages);
		pdfPagesRenderState = new Array(pdfDoc.numPages);
		for (var i = 0; i < pdfPagesRenderState.length; ++i) {
			pdfPagesRenderState[i] = RENDER_STATE_INIT;
		}
		if (lastPageNum != -1) {
			--lastPageNum;
			currentPageNum = lastPageNum;
			UpdatePageNumView();			
		}
		ParsePDFPage(0);
	});
}

function FakeLoading() { // 模拟加载进度，实际是在等待第一页渲染完成
	renderingPercent += Math.floor((100 - renderingPercent) * (0.5*Math.random()));
	var downloadPercent = document.getElementById("download_percent");
	downloadPercent.innerText = Math.floor(loadingPercent * percentWeight + renderingPercent * (1.0 - percentWeight)) + "%";
	if (firstPagePaint) {
		clearTimeout(loadingFailTimeoutId);
		return;
	} else {
		fakeLoadingTimeoutId = setTimeout(FakeLoading, fakeLoadingInterval);
	}
}

function ParsePDFPage(pageNum) {
	if (pageNum >= pdfDoc.numPages) {
		return;
	}
	pdfDoc.getPage(pageNum + 1) // PDF.js的页面编号从1开始
		.then(function(page) {
		pdfPages[pageNum] = page;
		pdfPagesRenderState[pageNum] = RENDER_STATE_PARSED;
		var pageWidth = page.view[2];
		pdfPagesRenderScale[pageNum] = Math.ceil((desiredWidth * pageScaleLimit) / pageWidth);
		numParsedPage = pageNum + 1;
		
		var container = CreatePageContainer(pageNum);
		document.body.appendChild(container);
		var margin = CreatePageMargin(pageNum);
		document.body.appendChild(margin);
		
		if (lastPageNum == -1) {
			if (pageNum == Math.min(pdfDoc.numPages - 1, 5)) {
				currentPageNum = 0;
				RenderPDFPages(currentPageNum);
			}
		} else {
			if (pageNum == Math.min(pdfDoc.numPages - 1, lastPageNum + 5)) {
				currentPageNum = lastPageNum;
				RenderPDFPages(currentPageNum);
			}
		}
		
		ParsePDFPage(pageNum + 1);
	});
}

function RenderPDFPages(startPageNum) {
	willRenderPageNumLow = Math.max(0, startPageNum - 2);
	willRenderPageNumHigh = Math.min(numParsedPage - 1, startPageNum + 4);
	var willRenderPageNumOrder = [
		startPageNum, startPageNum + 1, startPageNum - 1,
		startPageNum + 2, startPageNum + 3, startPageNum + 4, startPageNum - 2];

	RecyclePageContainers();
	RenderPDFPagesInternal(0);
	function RenderPDFPagesInternal(index) {
		if (index >= willRenderPageNumOrder.length) return;
		var pageNum = willRenderPageNumOrder[index];
		if (willRenderPageNumLow <= pageNum && pageNum <= willRenderPageNumHigh) {
			if (pdfPagesRenderState[pageNum] == RENDER_STATE_PARSED) {
				RenderPDFPage(pageNum, function (canvas) {
					onRenderFinish(pageNum, canvas);
					RenderPDFPagesInternal(index + 1);
				});
			} else {
				RenderPDFPagesInternal(index+1);
			}
		} else {
			if (0 <= pageNum && pageNum < numParsedPage && pdfPagesRenderState[pageNum] == RENDER_STATE_RENDERED) {
				ReleasePageContainer(pageNum);
				pdfPagesRenderState[pageNum] = RENDER_STATE_PARSED
			}
			RenderPDFPagesInternal(index+1);
		}
	}
	function onRenderFinish(pageNum, canvas) {
		if (willRenderPageNumLow <= pageNum && pageNum <= willRenderPageNumHigh) {
			AddCanvasToPageContainer(pageNum, canvas);
			pdfPagesRenderState[pageNum] = RENDER_STATE_RENDERED;
			if (!firstPagePaint && ((lastPageNum == -1 && pageNum == 0) || (pageNum == lastPageNum))) {
				var downloadProgressDiv = document.getElementById("download_progress_wrapper");
				downloadProgressDiv.parentNode.removeChild(downloadProgressDiv);
				setTimeout(function () {
					SignalWebEvent("PDF_FIRST_PAGE_PAINT");
				}, 500);
				firstPagePaint = true;
				canvas.scrollIntoView();
			}
		} else {
			ReleasePageContainer(pageNum);
			pdfPagesRenderState[pageNum] = RENDER_STATE_PARSED;
		}
	}

}

function RenderPDFPage(pageNum, onRenderFinishCallback) {
	pdfPagesRenderState[pageNum] = RENDER_STATE_RENDERING;
	var canvas = CreatePageCanvas(pageNum);
	var context = canvas.getContext("2d");
	var renderContext = {
		canvasContext : context,
		viewport : pdfPages[pageNum].getViewport(pdfPagesRenderScale[pageNum])
	};
	pdfPages[pageNum].render(renderContext).promise.then(function() {
		onRenderFinishCallback(canvas);
	});
}

function CreatePageContainer(pageNum) {
	var viewport = pdfPages[pageNum].getViewport(pdfPagesRenderScale[pageNum]);
	var container = document.createElement("div");
	container.id = "page_container_" + pageNum;
	container.className = "page_container";
	container.style.height = (viewport.height / viewport.width * desiredWidth) + "px";
	var hint = CreateRenderingHint();
	container.appendChild(hint);
	return container;
}

function CreateRenderingHint() {
	var hint = document.createElement("div");
	hint.className = "rendering_hint";
	hint.innerText = renderingHintText;
	return hint;
}

function CreatePageMargin(pageNum) {
	var margin = document.createElement("div");
	margin.id = "page_margin_" + pageNum;
	margin.className = "page_margin";
	return margin;
}

function CreatePageCanvas(pageNum) {
	var viewport = pdfPages[pageNum].getViewport(pdfPagesRenderScale[pageNum]);
	var canvas = document.createElement("canvas");
	canvas.id = "page_canvas_" + pageNum;
	canvas.className = "page_canvas";
	canvas.height = viewport.height;
	canvas.width = viewport.width;
	return canvas;
}

function AddCanvasToPageContainer(pageNum, canvas) {
	if (pageNum >= numParsedPage) return;
	var container = document.getElementById("page_container_" + pageNum);
	container.replaceChild(canvas, container.firstChild);
}

function RecyclePageContainers() {
	for (var pageNum = 0; pageNum < numParsedPage; ++pageNum) {
		if (willRenderPageNumLow <= pageNum && pageNum <= willRenderPageNumHigh) continue;
		if (pdfPagesRenderState[pageNum] == RENDER_STATE_RENDERED) {
			ReleasePageContainer(pageNum);
			pdfPagesRenderState[pageNum] = RENDER_STATE_PARSED
		}
	}
}

function ReleasePageContainer(pageNum) {
	if (pageNum >= numParsedPage) return;
	var container = document.getElementById("page_container_" + pageNum);
	var hint = CreateRenderingHint();
	container.replaceChild(hint, container.firstChild);
}

function FindCurrentPageNum() {
	for (var pageNum = 0; pageNum < numParsedPage; ++pageNum) {
		var container = document.getElementById("page_container_" + pageNum);
		var box = container.getBoundingClientRect();
	    if (box.bottom >= screen.height / 2  && box.top <= screen.height) {
			return pageNum;
		}
	}
	return 0;
}

function UpdatePageNumView() {
	SignalWebEvent("PAGE_NUM_UPDATE," + (currentPageNum + 1) + "," + pdfDoc.numPages)
}

function SignalWebEvent(msg) {
	prompt(msg);
}

var timeoutId = false;
function OnScroll() { 
	if (timeoutId) {
		clearTimeout(timeoutId);
		var pn = FindCurrentPageNum();
		if (pn != currentPageNum) {
			currentPageNum = pn;
			UpdatePageNumView();
		}
	}
	timeoutId = setTimeout(function() {
		var pn = FindCurrentPageNum();
		currentPageNum = pn;
		RenderPDFPages(currentPageNum);		
	}, 100);
}

window.onscroll = OnScroll;
